from typing import List

from dero.latex.logic.extract.extract import extract_objs_of_type_from_ambiguous_collection
from dero.latex.models.documentitem import DocumentItem

def extract_document_items_from_ambiguous_collection(collection) -> List[DocumentItem]:
    return extract_objs_of_type_from_ambiguous_collection(collection, DocumentItem)


from typing import Tuple
import os
from copy import deepcopy

from dero.manager.basemodels.file import ConfigFileBase
from dero.manager.imports.models.statements.container import ImportStatementContainer
from dero.manager.assignments.models.container import AssignmentStatementContainer

ImportsAndAssigns = Tuple[ImportStatementContainer, AssignmentStatementContainer]

class ConfigBase(dict):

    ##### Scaffolding functions or attributes. Need to override when subclassing  ####

    config_file_class = ConfigFileBase

    ##### Base class functions and attributes below. Shouldn't usually need to override in subclassing #####

    def __init__(self, d: dict=None, name: str=None, annotations: dict=None, imports: ImportStatementContainer = None,
                 _file: ConfigFileBase=None, begin_assignments: AssignmentStatementContainer=None, **kwargs):
        if d is None:
            d = {}
        super().__init__(d, **kwargs)

        if annotations is None:
            annotations = {}

        if imports is None:
            imports = ImportStatementContainer([])

        if begin_assignments is None:
            begin_assignments = AssignmentStatementContainer([])

        self.name = name
        self.annotations = annotations
        self.imports = imports
        self._file = _file
        self.begin_assignments = begin_assignments

    def __repr__(self):
        dict_repr = super().__repr__()
        class_name = self.__class__.__name__
        return f'<{class_name}(name={self.name}, {dict_repr})>'

    def __getattr__(self, attr):
        try:
            self[attr]
        except KeyError:
            raise AttributeError(attr)

    def __dir__(self):
        return self.keys()

    def update(self, d: dict=None, **kwargs):
        if d is None:
            d = {}
        super().update(d, **kwargs)

    def to_file(self, filepath: str):

        if self._file is None:
            output_file = self.config_file_class(filepath, name=self.name)
        else:
            # In case this is a new filepath for the same config, copy old file contents for use in new filepath
            output_file = deepcopy(self._file)
            output_file.filepath = filepath

        if os.path.exists(filepath):
            output_file.load() # load any existing config saved in the file, for preserving of user-saved inputs

        output_file.save(self)

    @classmethod
    def from_file(cls, filepath: str, name: str = None):
        file = cls.config_file_class(filepath, name=name)
        return file.load(cls)

    def as_imports_and_assignments(self) -> ImportsAndAssigns:
        assigns = AssignmentStatementContainer.from_dict_of_varnames_and_ast(self, self.annotations)

        return self.imports, self.begin_assignments + assigns

    def copy(self):
        return deepcopy(self)


import os
from functools import partial
from typing import List, Union

from .dropbox import filepath as dropbox_filepath

ListOfStrs = List[str]
StrOrListOfStrs = Union[ListOfStrs, str]

data_filepath = r'E:\Data'

def data_path(subpath: StrOrListOfStrs):
    return _os_join_list_or_str(data_filepath, subpath)

def data_path_func(subpath: StrOrListOfStrs):
    return path_func(subpath, data_path)

def dropbox_path(subpath: StrOrListOfStrs):
    return _os_join_list_or_str(dropbox_filepath, subpath)

def dropbox_path_func(subpath: StrOrListOfStrs):
    return path_func(subpath, dropbox_path)

def path_func(subpath: StrOrListOfStrs, path_func):
    new_path = path_func(subpath)
    new_path_func = partial(_os_join_list_or_str, new_path)
    return new_path_func

def _os_join_list_or_str(basepath: str, paths: StrOrListOfStrs) -> str:

    # If a list is passed, combine using operating system-specific path separator
    if isinstance(paths, list):
        paths = os.path.sep.join(paths)

    return os.path.join(basepath, paths)

#### Project paths

etf_project_path = dropbox_path_func(['UF', 'Andy', 'ETF Project'])
etf_path_func = lambda subpath: path_func(subpath, etf_project_path)


from dero.latex.df import df_to_pdf_and_move
from dero.latex.tools import date_time_move_latex, csv_to_raw_latex
from dero.latex.equations import latex_equations_to_pdf

from dero.latex.models.document import Document




"""
Pipeline Manager is a Python framework for flow-based programming and managing configuration. To use it, you register
your main functions, and config files are created automatically for them. It provides a way to run these functions
individually, in a list, in sections, or a combination thereof. Configuration has multiple layers, with the defaults
for the function being the lowest level config, which then gets overrided by project config, which gets overrided by
high-level section configs, then lower-level section configs, and finally by any local config updates in a script.
"""

from dero.manager.main import PipelineManager
from dero.manager.selector.models.selector import Selector
from dero.manager.data.models.merge import MergeOptions
from dero.manager.data.models.source import DataSource


from typing import Callable, Tuple
from functools import partial

from dero.mixins.repr import ReprMixin
from dero.manager.config.models.manager import ConfigManager, ActiveFunctionConfig
from dero.manager.pipelines.models.registrar import PipelineRegistrar, PipelineCollection
from dero.manager.logic.get import _get_public_name_or_special_name
from dero.manager.runner.models.interfaces import (
    StrOrListOfStrs,
    Result,
    Results,
    ResultOrResults
)
from dero.manager.pipelines.models.interfaces import (
    PipelineOrFunctionOrCollection,
    PipelineOrFunction
)
from dero.manager.sectionpath.sectionpath import SectionPath
from dero.manager.config.logic.write import dict_as_function_kwarg_str
from dero.manager.basemodels.pipeline import Pipeline
from dero.manager.views.object import ObjectView

class Runner(ReprMixin):
    repr_cols = ['_config', '_pipelines']


    def __init__(self, config: ConfigManager, pipelines: PipelineRegistrar):
        self._config = config
        self._pipelines = pipelines
        self._full_getattr = ''

    def __getattr__(self, item):
        # TODO: find way of doing this with fewer side effects
        # TODO: find a way to get this working for sections
        self._full_getattr += item
        try:
            func_or_collection = self._get_func_or_collection(self._full_getattr)
        except KeyError as e:
            self._full_getattr = '' # didn't find an item, must be incorrect path. reset total path
            raise e

        if isinstance(func_or_collection, PipelineCollection):
            # Got section, need to keep going. Return self
            self._full_getattr += '.' # add period to separate next section
            return self
        else:
            # Got function or Pipeline, return the function or pipeline itself
            configured_func_or_pipeline = self.get(self._full_getattr)
            self._full_getattr = ''  # found the item, reset for next time
            return configured_func_or_pipeline

    # TODO: this may work once __getattr__ works with sections
    # def __dir__(self):
    #     exposed_methods = [
    #         'run',
    #         'get',
    #     ]
    #
    #     if self._full_getattr != '':
    #         # Already into nested runner, need to pull appropriate collection
    #         try:
    #             func_or_collection = self._get_func_or_collection(self._full_getattr)
    #             pipeline_attrs = [attr for attr in dir(func_or_collection) if attr not in exposed_methods]
    #         except KeyError:
    #             pipeline_attrs = []
    #
    #     else:
    #         # Original runner. Just pull pipeline register attrs
    #         pipeline_attrs = [attr for attr in dir(self._pipelines) if attr not in exposed_methods]
    #
    #     return exposed_methods + pipeline_attrs


    def run(self, section_path_str_or_list: StrOrListOfStrs) -> ResultOrResults:
        """
        Use to run registered pipelines/functions/sections. Pass a single section path or a list
        of section paths. If a list is passed, the return value will also be a list, with each
        result item corresponding to the function at the same index.

        If a path for a section is passed rather than a path for a function or pipeline, a list
        of results will be returned for that section as well. Therefore, calling a high-level
        section will result in a nested list structure of results.

        Args:
            section_path_str_or_list: . separated name of path of function or section, or list thereof.
                similar to how a function would be imported. e.g. 'main.data.summarize.summary_func1'
                or when running multiple functions/sections, e.g.
                    ['main.data', 'main.analysis.reg.1']
                These sections/functions are based on the structure of your pipeline_dict

        Returns: result or list of results

        """
        if isinstance(section_path_str_or_list, str):
            # Running single function/pipeline
            return self._run(section_path_str_or_list)
        elif isinstance(section_path_str_or_list, list):
            return [self._run(section_path_str) for section_path_str in section_path_str_or_list]

    def _run(self, section_path_str: str) -> ResultOrResults:
        """
        Internal run function for running a single section path string. Handles both running
        sections and running individual functions/pipelines

        Args:
            section_path_str:

        Returns:

        """
        func_or_collection = self._get_func_or_collection(section_path_str)
        if isinstance(func_or_collection, PipelineCollection):
            return self._run_section(section_path_str)
        elif type(func_or_collection) is type and issubclass(func_or_collection, Pipeline):
            return self._run_one_pipeline(section_path_str)
        elif isinstance(func_or_collection, Callable):
            return self._run_one_func(section_path_str)
        else:
            raise ValueError(f'could not run section {section_path_str}. expected PipelineCollection or function,'
                             f'got {func_or_collection} of type {type(func_or_collection)}')


    def _run_section(self, section_path_str: str) -> Results:
        section: PipelineCollection = self._get_func_or_collection(section_path_str)
        results = []
        for section_or_object_view in section:

            # Get from object view if necessary
            if isinstance(section_or_object_view, ObjectView):
                section_or_callable = section_or_object_view.item
            else:
                section_or_callable = section_or_object_view

            # Get section path by which to call this item
            subsection_name = _get_public_name_or_special_name(section_or_callable, accept_output_names=False)
            subsection_path_str = SectionPath.join(section_path_str, subsection_name).path_str

            if isinstance(section_or_callable, PipelineCollection):
                # got another section within this section. recursively call run section
                results.append(self._run_section(subsection_path_str))
            elif type(section_or_callable) is type and issubclass(section_or_callable, Pipeline):
                # run pipeline
                results.append(self._run_one_pipeline(subsection_path_str))
            elif isinstance(section_or_callable, Callable):
                # run function
                results.append(self._run_one_func(subsection_path_str))
            else:
                raise ValueError(f'could not run section {subsection_path_str}. expected PipelineCollection or '
                                 f'function or Pipeline,'
                                 f'got {section_or_callable} of type {type(section_or_callable)}')

        return results


    def _run_one_func(self, section_path_str: str) -> Result:
        func, config_dict = self._get_func_and_config(section_path_str)

        print(f'Running function {section_path_str}({dict_as_function_kwarg_str(config_dict)})')
        result = func(**config_dict)
        print(f'Result:\n{result}\n')

        return result

    def _run_one_pipeline(self, section_path_str: str) -> Result:
        pipeline_class, config_dict = self._get_pipeline_and_config(section_path_str)

        # Construct new pipeline instance with config args
        configured_pipeline = pipeline_class(**config_dict)

        print(f'Running pipeline {configured_pipeline}({dict_as_function_kwarg_str(config_dict)})')
        result = configured_pipeline.execute()
        print(f'Result:\n{result}\n')

        return configured_pipeline

    def get(self, section_path_str: str) -> PipelineOrFunction:
        func_or_collection = self._get_func_or_collection(section_path_str)
        if isinstance(func_or_collection, PipelineCollection):
            # TODO: get sections
            raise NotImplementedError('have not implemented getting sections. works with run')
        elif type(func_or_collection) is type and issubclass(func_or_collection, Pipeline):
            # Got pipeline class
            return self._get_one_pipeline_with_config(section_path_str)
        elif isinstance(func_or_collection, Callable):
            return self._get_one_func_with_config(section_path_str)
        else:
            raise ValueError(f'could not get section {section_path_str}. expected PipelineCollection or function,'
                             f'got {func_or_collection} of type {type(func_or_collection)}')

    def _get_one_func_with_config(self, section_path_str: str) -> Callable:
        func, config_dict = self._get_func_and_config(section_path_str)

        full_func = partial(func, **config_dict)

        return full_func

    def _get_one_pipeline_with_config(self, section_path_str: str) -> Pipeline:
        pipeline_class, config_dict = self._get_pipeline_and_config(section_path_str)

        # Construct new pipeline instance with config args
        return pipeline_class(**config_dict)


    def _get_config(self, section_path_str: str) -> ActiveFunctionConfig:
        return self._config.get(section_path_str)

    def _get_func_or_collection(self, section_path_str: str) -> PipelineOrFunctionOrCollection:
        return self._pipelines.get(section_path_str)

    def _get_func_and_config(self, section_path_str: str) -> Tuple[Callable, dict]:
        config = self._get_config(section_path_str)
        func = self._get_func_or_collection(section_path_str)

        # Only pass items in config which are arguments of function
        config_dict = config.for_function(func)

        return func, config_dict

    def _get_pipeline_and_config(self, section_path_str: str) -> Tuple[type, dict]:
        config = self._get_config(section_path_str)
        # pipeline is an instance of the pipeline, without config
        pipeline = self._get_func_or_collection(section_path_str)

        # Only pass items in config which are arguments of function
        config_dict = config.for_function(pipeline.__init__)

        return pipeline, config_dict


import pandas as pd
from typing import Tuple, List, Union

InteractionTuple = Tuple[str, str]
InteractionTuples = List[InteractionTuple]

StrOrListOfStrs = Union[str, List[str]]

StrOrBool = Union[str, bool]
StrOrListOfStrsOrNone = Union[StrOrListOfStrs, None]




def _estimate_handling_robust_and_cluster(regdf, model, robust, cluster, **fit_kwargs):
    assert not (robust and cluster)  # need to pick one of robust or cluster

    if robust:
        return model.fit(cov_type='robust', **fit_kwargs)

    # TODO: linearmodels panel cluster
    if cluster:
        raise NotImplementedError('need to implement cluster for linearmodels panel models')

    return model.fit(**fit_kwargs)


from typing import List
import os

from dero.manager.basemodels.container import Container
from dero.mixins.repr import ReprMixin
from dero.manager.imports.models.statements.container import ImportStatementContainer

StrList = List[str]

scaffolding_error = NotImplementedError('must use DataCollection or PipelineCollection, not base class Collection')

class Collection(Container, ReprMixin):

    #### Scaffolding functions. These should be overridden by collection subclasses ###

    def _set_name_map(self):
        raise scaffolding_error

    def _output_config_files(self):
        raise scaffolding_error

    def _transform_item(self, item):
        """
        Is called on each item when adding items to collection. Should handle whether the item
        is an actual item or another collection. Must return the item or collection.

        If not overridden, will just store items as is.

        Returns: item or Collection

        """
        return item

    ### base functions. These probably do not need to be overridden by collection subclasses ###

    repr_cols = ['name', 'basepath', 'items']

    def __init__(self, basepath: str, items, name: str = None,
                 imports: ImportStatementContainer = None):
        self.basepath = basepath
        self.imports = imports
        self.items = self._transform_items(items)
        self.name = name


    def __getattr__(self, item):
        try:
            return self.name_dict[item]
        except KeyError:
            raise AttributeError(f'{item}')

    def __dir__(self):
        return self.name_dict.keys()

    @property
    def items(self):
        return self._items

    @items.setter
    def items(self, items):
        self._items = items
        self._set_name_map()  # need to recreate pipeline map when items change

    @classmethod
    def from_dict(cls, dict_: dict, basepath: str, name: str = None,
                  imports: ImportStatementContainer = None):
        items = []
        for section_name, dict_or_list in dict_.items():
            section_basepath = os.path.join(basepath, section_name)
            if isinstance(dict_or_list, dict):
                # Got another pipeline dict. Recursively process
                items.append(
                    cls.from_dict(
                        dict_or_list, basepath=section_basepath, name=section_name, imports=imports
                    )
                )
            elif isinstance(dict_or_list, list):
                # Got a list of functions or pipelines. Create a collection directly from items
                items.append(
                    cls.from_list(
                        dict_or_list, basepath=section_basepath, name=section_name, imports=imports
                    )
                )

        return cls(basepath=basepath, items=items, name=name)

    @classmethod
    def from_list(cls, list_: list, basepath: str, name: str = None,
                  imports: ImportStatementContainer = None):
        items = []
        for dict_or_item in list_:
            if isinstance(dict_or_item, dict):
                items.append(
                    cls.from_dict(
                        dict_or_item, basepath=basepath, name=name, imports=imports
                    )
                )
            else:
                # pipeline or function
                items.append(dict_or_item)

        return cls(basepath=basepath, items=items, name=name, imports=imports)

    def _transform_items(self, items):
        return [self._transform_item(item) for item in items]



from typing import Tuple, Union
import pandas as pd
import statsmodels.api as sm

from dero.reg.regtypes import StrOrListOfStrs, InteractionTuples, StrOrBool, StrOrListOfStrsOrNone
from dero.reg.lag.main import create_lagged_variables_return_yvars_xvars_interaction_tuples
from dero.reg.dataprep import (
    _set_interaction_tuples,
    _collect_all_variables_from_xvars_and_interaction_tuples,
    _drop_missings_df,
    _y_X_from_df,
    _set_fe
)

YvarXvars = Tuple[str, StrOrListOfStrs]
DummyDictOrNone = Union[dict, None]
DfYvarXvarsDummyDict = Tuple[pd.DataFrame, str, StrOrListOfStrs, DummyDictOrNone]
DfYvarXvarsLagvarsDummyDict = Tuple[pd.DataFrame, str, StrOrListOfStrs, StrOrListOfStrs, DummyDictOrNone]

def _create_reg_df_y_x_and_lag_vars(df: pd.DataFrame, yvar: str, xvars: StrOrListOfStrs,
                                    entity_var: str, time_var: str,
                                    cluster=False, cons=True, fe=None, interaction_tuples=None,
                                   num_lags=0, lag_variables='xvars', lag_period_var='Date', lag_id_var='TICKER',
                                    fill_method: str = 'ffill'
                                    ) -> DfYvarXvarsLagvarsDummyDict:

    # Handle lags
    df, reg_yvar, reg_xvars, interaction_tuples, lag_variables = create_lagged_variables_return_yvars_xvars_interaction_tuples(
        df, yvar, xvars,
        interaction_tuples=interaction_tuples,
        num_lags=num_lags,
        lag_variables=lag_variables,
        lag_period_var=lag_period_var,
        lag_id_var=lag_id_var,
        fill_method=fill_method
    )

    fe = _set_fe(fe)
    interaction_tuples = _set_interaction_tuples(interaction_tuples)
    regdf, y, X, dummy_cols_dict = _get_reg_df_y_x(df, reg_yvar, reg_xvars, entity_var, time_var, cluster, cons, fe, interaction_tuples)
    return regdf, y, X, lag_variables, dummy_cols_dict

def _get_reg_df_y_x(df: pd.DataFrame, yvar: str, xvars: StrOrListOfStrs, entity_var: str, time_var: str,
                    cluster: StrOrBool,
                    cons: bool, fe: StrOrListOfStrsOrNone, interaction_tuples: InteractionTuples) -> DfYvarXvarsDummyDict:
    all_xvars = _collect_all_variables_from_xvars_and_interaction_tuples(xvars, interaction_tuples)
    regdf = _drop_missings_df(df, yvar, all_xvars, cluster, fe)
    regdf = regdf.set_index([entity_var, time_var])
    y, X, dummy_cols_dict = _y_X_from_df(regdf, yvar, xvars, cons, fe, interaction_tuples)

    return regdf, y, X, dummy_cols_dict


from typing import Callable, Any, List
import inspect

from dero.manager.basemodels.config import ConfigBase
from dero.manager.imports.logic.load.func import function_args_as_dict
from dero.manager.basemodels.pipeline import Pipeline
from dero.manager.config.models.file import FunctionConfigFile

class FunctionConfig(ConfigBase):

    config_file_class = FunctionConfigFile

    def for_function(self, func: Callable) -> dict:
        """
        Strips out items of config which are not applicable to function. Returns dictionary
        of config items for passing to the function.

        Args:
            func: func for which to filter out config items

        Returns: dict, applicable config for func
        """
        # Only pass items in config which are arguments of function
        func_kwargs = function_args_as_dict(func)
        return {key: value for key, value in self.items() if key in func_kwargs}



from dero.manager.basemodels.config import ConfigBase
from dero.manager.data.models.source import DataSource, DataType
from dero.manager.data.models.file import DataConfigFile, ConfigFileBase
from dero.manager.imports.models.statements.container import ImportStatementContainer
from dero.manager.assignments.models.container import AssignmentStatementContainer
from dero.manager.io.file.load.parsers.fromdict import extract_dict_from_ast_dict_or_dict_constructor
from dero.manager.data.models.astitems import ast_str, ast_dict_constructor


class DataConfig(ConfigBase):
    config_file_class = DataConfigFile

    def __init__(self, d: dict = None, name: str = None, annotations: dict = None,
                 imports: ImportStatementContainer = None,
                 _file: ConfigFileBase = None, begin_assignments: AssignmentStatementContainer = None,
                 active_config_dict: dict = None, **kwargs):

        if active_config_dict is None:
            active_config_dict = {}

        super().__init__(
            d=d,
            name=name,
            annotations=annotations,
            imports=imports,
            _file=_file,
            begin_assignments=begin_assignments,
            **kwargs
        )

        self.active_config_dict = active_config_dict


    @classmethod
    def from_source(cls, data_source: DataSource, name: str=None, imports: ImportStatementContainer = None):
        # Initialize a blank config dictionary
        config_dict = DataSource._scaffold_dict.copy()

        if data_source.loader_func_kwargs == {}:
            # If empty dict, must be where file doesn't exist, and so loader_func_kwargs were set to default
            # Need to convert to ast version
            data_source.loader_func_kwargs = ast_dict_constructor

        # Fill blank config dict
        for config_attr in config_dict:
            if not _source_attr_is_none(data_source, config_attr):
                config_dict[config_attr] = getattr(data_source, config_attr)

        # Special handling for name, which will be set even before creating file
        if isinstance(config_dict['name'], str):
            config_dict['name'] = ast_str(config_dict['name'])  # convert str to ast


        return cls(config_dict, name=name, imports=imports)

def _source_attr_is_none(source: DataSource, source_attr: str) -> bool:
    source_value = getattr(source, source_attr, None)
    if source_value == None:  # using equals to call __eq__ method of DataType
        return True

    return False




from dero.latex.models.documentitem import DocumentItem
from dero.latex.models.title import Title
from dero.latex.models.author import Author
from dero.latex.models.date import Date
from dero.latex.texgen import _maketitle_str

class TitlePage(DocumentItem):

    def __init__(self, title=None, author=None, date=None):
        from dero.latex.logic.builder import _build

        self.contents = [
            Title(title) if title is not None else None,
            Author(author) if author is not None else None,
            Date(date) if date is not None else Date(),
            _maketitle_str()
        ]

        self._output = _build(self.contents)

    def __str__(self):
        return self._output


from typing import Union

from dero.manager.data.models.source import DataSource

StrOrDataSource = Union[str, DataSource]

def convert_to_data_source_if_necessary(item: StrOrDataSource) -> DataSource:
    from dero.manager.data.models.collection import DataCollection
    if isinstance(item, (DataSource, DataCollection)):
        return item

    return DataSource(name=item)


from dero.latex.logic.tools import _readable_repr


class ReprMixin:
    repr_cols = []

    def __repr__(self):
        if self.repr_cols:
            repr_col_strs = [f'{col}={getattr(self, col, None).__repr__()}' for col in self.repr_cols]
            repr_col_str = f'({", ".join(repr_col_strs)})'
        else:
            repr_col_str = ''
        return f'<{type(self).__name__}{repr_col_str}>'

    @property
    def readable_repr(self):
        return show_contents(self)


def show_contents(obj):
    """
    Used to view what's inside a latex table object

    >>>import dero.latex.table as lt
    >>>dt = lt.DataTable.from_df(some_df)
    >>>show_contents(dt)
    :param obj:
    :return:
    """
    print(_readable_repr(repr(obj)))


from typing import List, Union
import os

from dero.manager.basemodels.collection import Collection
from dero.manager.data.models.source import DataSource
from dero.manager.data.logic.convert import convert_to_data_source_if_necessary
from dero.manager.logic.get import _get_public_name_or_special_name
from dero.manager.data.models.config import DataConfig


SourceOrCollection = Union[DataSource, 'DataCollection']

class DataCollection(Collection):

    def _set_name_map(self) -> None:
        source_map = {}
        for source_or_collection in self:
            source_or_collection: SourceOrCollection
            source_name = _get_public_name_or_special_name(source_or_collection)
            source_map[source_name] = source_or_collection
        self.name_dict = source_map

    def _transform_item(self, item):
        """
        Is called on each item when adding items to collection. Should handle whether the item
        is an actual item or another collection. Must return the item or collection.

        If not overridden, will just store items as is.

        Returns: item or Collection

        """
        return convert_to_data_source_if_necessary(item)

    def _output_config_files(self) -> None:
        if not os.path.exists(self.basepath):
            os.makedirs(self.basepath)

        [self._output_config_file(item) for item in self]

    def _output_config_file(self, item: SourceOrCollection) -> None:
        if isinstance(item, DataCollection):
            # if collection, recursively call creating config files
            return item._output_config_files()

        # Dealing with DataSource
        item_name = _get_public_name_or_special_name(item)
        item_filepath = os.path.join(self.basepath, item_name + '.py')

        if os.path.exists(item_filepath):
            file_existed = True
            # if config file already exists, load confguration from file, use to update file defaults
            existing_config = DataConfig.from_file(item_filepath, item_name)
            existing_imports = existing_config._file.interface.imports
            # Here using the ast config, for the purpose of writing to file
            file_config_item = item.copy()
            file_config_item.apply_config(existing_config)
            # also use to update DataSource object in memory. Here use the actual objects instead of ast
            item.apply_config(existing_config.active_config_dict)
        else:
            file_existed = False
            existing_imports = None
            file_config_item = item

        item_config = DataConfig.from_source(file_config_item, imports=existing_imports)
        item_config.to_file(item_filepath)

        if not file_existed:
            # If this was a new output, we now need to load again to get the object representation
            # instead of just the ast representation
            existing_config = DataConfig.from_file(item_filepath, item_name)
            item.apply_config(existing_config.active_config_dict)


import sys
import os
import traceback
import pdb
from typing import TYPE_CHECKING, Union, List
if TYPE_CHECKING:
    from dero.manager.selector.models.itemview import ItemView
    StrOrView = Union[str, ItemView]
    RunnerArgs = Union[str, List[str], ItemView, List[ItemView]]

from dero.manager.config.models.manager import ConfigManager
from dero.manager.pipelines.models.registrar import PipelineRegistrar
from dero.manager.pipelines.models.dictfile import PipelineDictFile
from dero.manager.data.models.registrar import DataRegistrar
from dero.manager.data.models.dictfile import DataDictFile
from dero.manager.runner.models.runner import Runner, ResultOrResults
from dero.manager.imports.models.tracker import ImportTracker
from dero.manager.sectionpath.sectionpath import SectionPath
from dero.manager.exceptions.pipelinemanager import PipelineManagerNotLoadedException


class PipelineManager:
    """
    Main class for managing flow-based programming and configuration.
    """

    def __init__(self, pipeline_dict_path: str, data_dict_path: str, basepath: str, name: str='project',
                 auto_pdb: bool=False):
        self.pipeline_dict_path = pipeline_dict_path
        self.data_dict_path = data_dict_path
        self.basepath = basepath
        self.sources_basepath = os.path.join(basepath, 'sources')
        self.name = name
        self.auto_pdb = auto_pdb


    def __getattr__(self, item):

        if item in ('runner', 'sources'):
            # must not be defined yet
            raise PipelineManagerNotLoadedException('call PipelineManager.load() before accessing '
                                                    'functions or data sources')

        # Must be getting function
        return getattr(self.runner, item)

    def __dir__(self):
        exposed_methods = [
            'run',
            'get',
            'load',
            'reload'
        ]
        exposed_attrs = ['sources', 'name']
        exposed = exposed_methods + exposed_attrs
        if hasattr(self, 'register'):
            skip_methods = [
                'scaffold_config',
                'basepath',
            ]
            register_attrs = [attr for attr in dir(self.register) if attr not in exposed + skip_methods]
            exposed += register_attrs
        return exposed

    def run(self, section_path_str_or_list: 'RunnerArgs') -> ResultOrResults:
        """
        Use to run registered pipelines/functions/sections. Pass a single section path or a list
        of section paths. If a list is passed, the return value will also be a list, with each
        result item corresponding to the function at the same index.

        If a path for a section is passed rather than a path for a function or pipeline, a list
        of results will be returned for that section as well. Therefore, calling a high-level
        section will result in a nested list structure of results.

        Args:
            section_path_str_or_list: . separated name of path of function or section, or list thereof.
                similar to how a function would be imported. e.g. 'main.data.summarize.summary_func1'
                or when running multiple functions/sections, e.g.
                    ['main.data', 'main.analysis.reg.1']
                These sections/functions are based on the structure of your pipeline_dict

        Returns: result or list of results

        """
        section_path_str_or_list = self._convert_list_or_single_item_view_or_str_to_strs(section_path_str_or_list)

        if not self.auto_pdb:
            return self.runner.run(section_path_str_or_list)

        # auto pdb
        try:
            return self.runner.run(section_path_str_or_list)
        except:
            traceback.print_exc()
            pdb.post_mortem()


    # TODO: multiple section path strs
    def get(self, section_path_str_or_view: 'StrOrView'):
        section_path_str = self._get_section_path_str_from_section_path_str_or_view(section_path_str_or_view)
        section_path = SectionPath(section_path_str)

        if section_path[0] == 'sources':
            return self.sources.get(section_path_str)
        else:
            return self.runner.get(section_path_str)

    def reload(self) -> None:
        """
        Useful for getting file system changes without having to start a new Python session. Also resets
        any locally defined configuration.

        Reloads functions from pipeline dict, scaffolds config files for any new functions,
        and updates configuration from files.

        Returns: None

        """
        self._wipe_loaded_modules()
        self.load()

    def load(self) -> None:
        """
        Wrapper to track imported modules so that can reimport them upon reloading
        """
        self._import_tracker = ImportTracker()

        try:
            self._load_pipeline_config_and_runner()
        except Exception as e:
            # Reset loaded modules from import, completely canceling load so it can be tried again
            self._loaded_modules = self._import_tracker.imported_modules
            self._wipe_loaded_modules()
            self._loaded_modules = [] # modules have been wiped, must be sure they won't be wiped again
            raise e

        self._loaded_modules = self._import_tracker.imported_modules

    def _load_pipeline_config_and_runner(self) -> None:
        """
        External logic for main initialization
        Returns:

        """
        # Load dynamically instead of passing dict to ensure modules are loaded into sys now
        pipeline_dict_file = PipelineDictFile(self.pipeline_dict_path, name='pipeline_dict')
        pipeline_dict = pipeline_dict_file.load()
        data_dict_file = DataDictFile(self.data_dict_path, name='data_dict')
        data_dict = data_dict_file.load()

        self.register = PipelineRegistrar.from_dict(
            pipeline_dict,
            basepath=self.basepath,
            name=self.name,
            imports=pipeline_dict_file.interface.imports
        )
        self.register.scaffold_config()

        self.config = ConfigManager(self.basepath)
        self.config.load()

        self.sources = DataRegistrar.from_dict(
            data_dict,
            basepath=self.sources_basepath,
            name=self.name,
            imports=data_dict_file.interface.imports
        )
        self.sources.scaffold_config()

        self.runner = Runner(config=self.config, pipelines=self.register)

    def _wipe_loaded_modules(self):
        [sys.modules.pop(module) for module in self._loaded_modules]

    def _convert_list_or_single_item_view_or_str_to_strs(self, run_list: 'RunnerArgs') -> Union[str, List[str]]:
        if isinstance(run_list, list):
            return [self._get_section_path_str_from_section_path_str_or_view(run_arg) for run_arg in run_list]
        else:
            return self._get_section_path_str_from_section_path_str_or_view(run_list)

    def _get_section_path_str_from_section_path_str_or_view(self,section_path_str_or_view: 'StrOrView') -> str:
        from dero.manager.selector.models.itemview import _is_item_view
        if _is_item_view(section_path_str_or_view):
            # ItemView will have PipelineManager.name as first section, must strip
            section_path = SectionPath(section_path_str_or_view.section_path_str)
            relative_section_path = SectionPath.from_section_str_list(section_path[1:])
            return relative_section_path.path_str
        elif isinstance(section_path_str_or_view, str):
            return section_path_str_or_view
        else:
            raise ValueError(f'expected str or ItemView. Got {section_path_str_or_view} of '
                             f'type {type(section_path_str_or_view)}')



from functools import partial
from typing import Union, Sequence, List, Callable
import datetime
from copy import deepcopy

from dero.manager.basemodels.pipeline import Pipeline
from dero.manager.data.models.source import DataSource
from dero.manager.data.models.merge import DataMerge, MergeOptions, LastMergeFinishedException
from dero.manager.selector.models.itemview import ItemView, _is_item_view

DataSourceOrPipeline = Union[DataSource, 'DataPipeline']
DataSourcesOrPipelines = Sequence[DataSourceOrPipeline]
MergeOptionsList = Sequence[MergeOptions]
DataMerges = List[DataMerge]


class DataPipeline(Pipeline):

    def __init__(self, data_sources: DataSourcesOrPipelines=None, merge_options_list: MergeOptionsList=None,
                 outpath=None, post_merge_cleanup_func=None, name: str=None, cleanup_kwargs: dict=None):

        if cleanup_kwargs is None:
            cleanup_kwargs = {}

        self.data_sources = data_sources
        self.merge_options_list = merge_options_list
        self._merge_index = 0
        self._set_cleanup_func(post_merge_cleanup_func, **cleanup_kwargs)
        self.outpath = outpath
        self.name = name
        self.cleanup_kwargs = cleanup_kwargs

    def execute(self):
        while True:
            try:
                self.next_merge()
            except LastMergeFinishedException:
                break

        if self.has_post_merge_cleanup_func:
            self.df = self.post_merge_cleanup_func(self.df)
        self.output()

        return self.df

    def next_merge(self):
        # On first merge, set df
        if self._merge_index == 0:
            self._set_df_from_first_merge()

        self._merge()

    def output(self, outpath=None):
        if outpath:
            self._output(outpath)
        elif self.outpath:
            self._output(self.outpath)

    def summary(self, *summary_args, summary_method: str=None, summary_function: Callable=None,
                             summary_attr: str=None, **summary_method_kwargs):
        for merge in self.merges:
            merge.summary(
                *summary_args,
                summary_method=summary_method,
                summary_function=summary_function,
                summary_attr=summary_attr,
                **summary_method_kwargs
            )

    def describe(self):
        for merge in self.merges:
            merge.describe()

    def _output(self, outpath=None):
        self.df.to_csv(outpath, index=False, encoding='utf8')

    def _merge(self):
        try:
            print(f'Now running merge {self._merge_index + 1}: {self.merges[self._merge_index]}')
        except IndexError:
            raise LastMergeFinishedException

        self.merges[self._merge_index].merge()

        # Set current df to result of merge
        self.df = self.merges[self._merge_index].result.df

        self._merge_index += 1

        # TODO: add output considering path in merge options

    @property
    def merges(self):
        try:
            return self._merges
        except AttributeError:
            self._set_merges()

        return self._merges

    # Following properties exist to recreate merges if data sources or merge options are overridden
    # by user

    @property
    def data_sources(self):
        return self._data_sources

    @data_sources.setter
    def data_sources(self, data_sources: DataSourcesOrPipelines):
        self._data_sources = data_sources
        # only set merges if previously set. otherwise no need to worry about updating cached result
        if hasattr(self, '_merges'):
            self._set_merges()

    @property
    def merge_options_list(self):
        return self._merge_options_list

    @merge_options_list.setter
    def merge_options_list(self, merge_options_list: MergeOptionsList):
        self._merge_options_list = merge_options_list
        # only set merges if previously set. otherwise no need to worry about updating cached result
        if hasattr(self, '_merges'):
            self._set_merges()

    def _set_merges(self):
        self._touch_data_sources()
        self._merges = self._create_merges(self.data_sources, self.merge_options_list)

    def _create_merges(self, data_sources: DataSourcesOrPipelines, merge_options_list: MergeOptionsList):
        merges = _get_merges(data_sources[0], data_sources[1], merge_options_list[0])
        if len(merge_options_list) == 1:
            return merges

        for i, merge_options in enumerate(merge_options_list[1:]):
            merges += _get_merges(merges[-1].result, data_sources[i + 2], merge_options)

        return merges

    def _set_df_from_first_merge(self):
        self.df = self.merges[0].data_sources[0].df

    def _set_cleanup_func(self, post_merge_cleanup_func, **cleanup_kwargs):
        if post_merge_cleanup_func is not None:
            self.has_post_merge_cleanup_func = True
            self.post_merge_cleanup_func = partial(post_merge_cleanup_func, **cleanup_kwargs)
        else:
            self.has_post_merge_cleanup_func = False

    @property
    def last_modified(self):
        self._touch_data_sources()
        return max([source.last_modified for source in self.data_sources])

    @property
    def source_last_modified(self):
        most_recent_time = datetime.datetime(1900, 1, 1)
        for i, source in enumerate(self.data_sources):
            if source.last_modified > most_recent_time:
                most_recent_time = source.last_modified
                most_recent_index = i

        return self.data_sources[most_recent_index]

    def copy(self):
        return deepcopy(self)

    def _touch_data_sources(self):
        """
        Data sources may be passed using dero.manager.Selector, in which case they
        are dero.manager.selector.models.itemview.ItemView objects. _get_merges uses isinstance, which will
        return ItemView, and so won't work correctly. By accessing the .item proprty of the ItemView,
        we get the original item back
        Returns:

        """
        real_data_sources = []
        for data_source_or_view in self.data_sources:
            if _is_item_view(data_source_or_view):
                data_source_or_view: ItemView
                real_data_sources.append(data_source_or_view.item)
            else:
                real_data_sources.append(data_source_or_view)
        self.data_sources = real_data_sources


def _get_merges(data_source_1: DataSourceOrPipeline, data_source_2: DataSourceOrPipeline,
                merge_options: MergeOptions) -> DataMerges:
    """
    Creates a list of DataMerge objects from a paring of two DataSource objects, a DataSource and a DataPipeline,
    or two DataPipeline objects.
    :param data_source_1: DataSource or DataPipeline
    :param data_source_2: DataSource or DataPipeline
    :param merge_options: MergeOptions
    :return: list of DataMerge objects
    """
    merges = []
    final_merge_sources = []
    # Add any pipeline merges first, as the results from the pipeline must be ready before we can merge the results
    # to other data sources or pipeline results
    if _is_data_pipeline(data_source_1):

        # TODO: implement cleanup funcs in pipelines of pipelines
        if data_source_1.has_post_merge_cleanup_func:
            raise NotImplementedError(f'no handling yet for post merge cleanup function in {data_source_1}')

        merges += data_source_1.merges
        pipeline_1_result = data_source_1.merges[-1].result
        final_merge_sources.append(pipeline_1_result) # result of first pipeline will be first source in final merge
    if _is_data_pipeline(data_source_2):

        # TODO: implement cleanup funcs in pipelines of pipelines
        if data_source_2.has_post_merge_cleanup_func:
            raise NotImplementedError(f'no handling yet for post merge cleanup function in {data_source_2}')

        merges += data_source_2.merges
        pipeline_2_result = data_source_2.merges[-1].result # result of second pipeline will be second source in final merge

    if not _is_data_pipeline(data_source_1):
        final_merge_sources.append(data_source_1)

    # Now final merge source 1 is filled, may add 2
    if _is_data_pipeline(data_source_2):
        final_merge_sources.append(pipeline_2_result)
    elif not _is_data_pipeline(data_source_2):
        final_merge_sources.append(data_source_2)

    # Add last (or only) merge
    merges.append(DataMerge(final_merge_sources, merge_options))

    return merges


def _is_data_pipeline(obj) -> bool:
    return hasattr(obj, 'data_sources') and hasattr(obj, 'merge_options_list')




from typing import List
import ast
import warnings

from dero.manager.io.func.load.config import function_args_as_arg_and_annotation_dict
from dero.manager.io.file.load.parsers.extname import extract_external_name_from_assign_value
from dero.manager.imports.models.statements.container import ImportStatementContainer, ObjectImportStatement
from dero.manager.assignments.models.container import AssignmentStatementContainer
from dero.manager.logic.inspect import _is_str_matching_builtin_type

def extract_import_statements_from_function_args_imports_and_assigns(args: ast.arguments,
                                                                     imports: ImportStatementContainer,
                                                                     assigns: AssignmentStatementContainer,
                                                                     current_module_section_path_str: str,
                                                                     ) -> ImportStatementContainer:
    defaults_dict, annotation_dict = function_args_as_arg_and_annotation_dict(args)

    # Extract external names. These are the root names (not attrs, the base of attrs) for anything that is not builtin
    external_names = _unique_external_names_from_default_dict_and_annotation_dict(defaults_dict, annotation_dict)

    out_imports = ImportStatementContainer([])

    # External names may be either due to assignment or due to imports
    # Handle names due to assignment
    for name in external_names:
        if assigns.contains_varname(name):
            # External name is due to assignment in this module. Create import from this module
            out_imports.append(ObjectImportStatement.from_str(f'from {current_module_section_path_str} import {name}'))

    # Handle names which are imported into this file
    for name in external_names:
        import_or_none = imports.get_import_for_module_or_obj_name(name)
        if import_or_none is not None:
            out_imports.append(import_or_none)

    # Sanity check, did we find all the imports?
    n_external = len(external_names)
    n_imports = len(out_imports)
    if n_external != n_imports:
        warnings.warn(f'had {n_external} external names from function definition, only '
                      f'found {n_imports} imports. May be missing imports')

    return out_imports

def _unique_external_names_from_default_dict_and_annotation_dict(default_dict: dict,
                                                                 annotation_dict: dict) -> List[str]:
    names = _unique_external_names_from_assignment_dict(default_dict)
    names += _unique_external_names_from_annotation_dict(annotation_dict)

    return list(set(names))

def _unique_external_names_from_multiple_assignment_dicts(*assignment_dicts) -> List[str]:
    all_names = []
    for assignment_dict in assignment_dicts:
        all_names += _unique_external_names_from_assignment_dict(assignment_dict)
    return list(set(all_names))

def _unique_external_names_from_assignment_dict(assignment_dict: dict) -> List[str]:
    names = [extract_external_name_from_assign_value(value) for key, value in assignment_dict.items()]
    names = [name for name in names if name is not None]
    return list(set(names))

def _unique_external_names_from_annotation_dict(annotation_dict: dict) -> List[str]:
    type_str_names = _extract_unique_type_str_names_from_annotation_dict(annotation_dict)
    return [type_str for type_str in type_str_names if not _is_str_matching_builtin_type(type_str)]

def _extract_unique_type_str_names_from_annotation_dict(annotation_dict: dict) -> List[str]:
    names = []
    for value in annotation_dict.values():
        if isinstance(value, ast.Name):
            names.append(value.id) # handles most types
        elif isinstance(value, ast.Subscript):
            # e.g.: List[str]
            names.append(value.value.id) # e.g., gets List portion
            names.append(value.slice.value.id) # e.g., gets str portion
        else:
            raise NotImplementedError(f'no handling for {value} of type {type(value)}')



    return list(set(names))


from copy import deepcopy
import pandas as pd
from functools import partial
import os
import warnings
import datetime
from typing import Callable, TYPE_CHECKING, List

from dero.ext_pandas.optimize.load import read_file as read_file_into_df

from dero.manager.data.models.type import DataType
from dero.manager.data.models.astitems import ast_none

if TYPE_CHECKING:
    from dero.manager.data.models.pipeline import DataPipeline

class DataSource:
    _scaffold_dict = {
        'name': ast_none,
        'data_type': ast_none,
        'location': ast_none,
        'loader_func': ast_none,
        'pipeline': ast_none,
        'tags': ast_none
    }

    # TODO: scaffold annotations

    def __init__(self, location: str =None, df: pd.DataFrame =None, pipeline: 'DataPipeline' =None,
                 name: str =None, data_type: str =None, tags: List[str]=None,
                 loader_func: Callable =None, loader_func_kwargs: dict=None):

        if loader_func_kwargs is None:
            loader_func_kwargs = {}

        self._check_inputs(location, df)
        self.location = location
        self.name = name
        self.data_type = data_type
        self.tags = tags # TODO: better handling for tags
        self.loader_func = loader_func
        self.pipeline = pipeline
        self.loader_func_kwargs = loader_func_kwargs
        self._df = df
        self.name_type = f'{name} {self.data_type}'

    def apply_config(self, config) -> None:
        from dero.manager.data.models.config import DataConfig
        config: DataConfig

        for config_attr, config_item in config.items():
            # Skip irrelevant items
            if hasattr(self, config_attr):
                setattr(self, config_attr, config_item)

    @property
    def df(self):
        if self._df is None:
            self._df = self._load()
        return self._df

    @property
    def data_type(self):
        return self._type

    @data_type.setter
    def data_type(self, dtype):
        self._type = DataType(dtype)

    @df.setter
    def df(self, df):
        self._df = df

    @property
    def last_modified(self):
        if self.location is None:
            # No location. Setting last modified time as a long time ago, so will trigger pipeline instead
            return datetime.datetime(1899, 1, 1)

        return datetime.datetime.fromtimestamp(os.path.getmtime(self.location))

    def _load(self):
        self._touch_pipeline()
        if not hasattr(self, 'data_loader'):
            self._set_data_loader(data_loader=self.loader_func, pipeline=self.pipeline, **self.loader_func_kwargs)
        return self.data_loader()

    def output(self, **to_csv_kwargs):
        self.df.to_csv(self.location, **to_csv_kwargs)

    def _check_inputs(self, filepath, df):
        pass
        # assert not (filepath is None) and (df is None)

    def _set_data_loader(self, data_loader: Callable =None, pipeline: 'DataPipeline' =None, **loader_func_kwargs):
        run_pipeline = False
        if pipeline is not None:
            # if a source in the pipeline to create this data source was modified more recently than this data source
            # note: if there is no location, will always enter the next block, as last modified time will set
            # to a long time ago
            if pipeline.last_modified > self.last_modified:
                # a prior source used to construct this data source has changed. need to re run pipeline
                recent_source = pipeline.source_last_modified
                warnings.warn(f'''data source {recent_source} was modified at {recent_source.last_modified}.

                this data source {self} was modified at {self.last_modified}.

                to get new changes, will load this data source through pipeline rather than from file.''')

                run_pipeline = True
            # otherwise, don't need to worry about pipeline, continue handling

        if self.location is None:
            # no location or pipeline, so accessing df will return empty dataframe
            loader = pd.DataFrame
        else:
            if data_loader is None:
                # TODO: determine filetype and use proper loader
                loader = partial(read_file_into_df, self.location, **loader_func_kwargs)
            else:
                loader = partial(data_loader, self.location, **loader_func_kwargs)

        # If necessary, run pipeline before loading
        # Still necessary to use loader as may be transforming the data
        if run_pipeline:
            def run_pipeline_then_load(pipeline):
                pipeline.execute() # outputs to file
                return loader()
            self.data_loader = partial(run_pipeline_then_load, self.pipeline)
        else:
            self.data_loader = loader


    def _touch_pipeline(self):
        """
        Pipeline may be passed using dero.manager.Selector, in which case it is
        a dero.manager.selector.models.itemview.ItemView object. _set_data_loader accesses a property of
        the pipeline before it's configured, and so won't work correctly. By accessing the .item proprty of the ItemView,
        we get the original item back
        Returns:

        """
        from dero.manager.selector.models.itemview import _is_item_view

        if _is_item_view(self.pipeline):
            self.pipeline = self.pipeline.item

    def copy(self):
        return deepcopy(self)

    def __repr__(self):
        return f'<DataSource(name={self.name}, type={self.data_type.name})>'


from dero.reg.dataprep import _create_reg_df_y_x_and_dummies, _estimate_handling_robust_and_cluster
from dero.reg.models import get_model_class_by_string, _is_probit_str, _is_logit_str


def reg(df, yvar, xvars, robust=True, cluster=False, cons=True, fe=None, interaction_tuples=None,
        num_lags=0, lag_variables='xvars', lag_period_var='Date', lag_id_var='TICKER',
        lag_fill_method: str='ffill',
        model_type='OLS'):
    """
    Returns a fitted regression. Takes df, produces a regression df with no missing among needed
    variables, and fits a regression model. If robust is specified, uses heteroskedasticity-
    robust standard errors. If cluster is specified, calculated clustered standard errors
    by the given variable.

    Note: only specify at most one of robust and cluster.

    Required inputs:
    df: pandas dataframe containing regression data
    yvar: str, column name of outcome y variable
    xvars: list of strs, column names of x variables for regression

    Optional inputs:
    robust: bool, set to True to use heterskedasticity-robust standard errors
    cluster: False or str, set to a column name to calculate standard errors within clusters
             given by unique values of given column name
    cons: bool, set to False to not include a constant in the regression
    fe: None or str or list of strs. If a str or list of strs is passed, uses these categorical
    variables to construct dummies for fixed effects.
    interaction_tuples: tuple or list of tuples of column names to interact and include as xvars
    num_lags: int, Number of periods to lag variables. Setting to other than 0 will activate lags
    lag_variables: 'all', 'xvars', or list of strs of names of columns to lag for regressions.
    lag_period_var: str, only used if lag_variables is not None. name of column which
                    contains period variable for lagging
    lag_id_var: str, only used if lag_variables is not None. name of column which
                    contains identifier variable for lagging
    lag_fill_method: str, 'ffill' or 'bfill' for which method to use to fill in missing rows when
                     creating lag variables. See pandas.DataFrame.fillna for more details
    model_type: str, 'OLS', 'probit', or 'logit' for type of model

    Returns:
    If fe=None, returns statsmodels regression result
    if fe is not None, returns a tuple of (statsmodels regression result, dummy_cols_dict)
    """
    regdf, y, X, dummy_cols_dict, lag_variables = _create_reg_df_y_x_and_dummies(df, yvar, xvars, cluster=cluster,
                                                                                 cons=cons, fe=fe,
                                                                                 interaction_tuples=interaction_tuples, num_lags=num_lags,
                                                                                 lag_variables=lag_variables, lag_period_var=lag_period_var,
                                                                                 lag_id_var=lag_id_var,
                                                                                 fill_method=lag_fill_method)

    ModelClass = get_model_class_by_string(model_type)
    mod = ModelClass(y, X)

    if _is_probit_str(model_type) or _is_logit_str(model_type):
        fit_kwargs = dict(
            method='bfgs',
            maxiter=100
        )
    else:
        fit_kwargs = {}

    result = _estimate_handling_robust_and_cluster(regdf, mod, robust, cluster, **fit_kwargs)

    # Only return dummy_cols_dict when fe is active
    if fe is not None:
        return result, dummy_cols_dict
    else:
        return result





import dero.latex.table as lt
from dero.data.typing import DfDict, StrOrNone

def missing_detail_df_dict_to_table_and_output(df_dict: DfDict, row_byvar, col_byvar: str,
                                               id_var: str,
                                               count_with_missings_var: str,
                                               missing_tolerance: int,
                                               missing_display_str: str = 'Missing',
                                               extra_caption: str='', extra_below_text: str='',
                                               outfolder: StrOrNone=None) -> lt.Table:

    caption = f'Data Gap Analysis - {missing_display_str} {count_with_missings_var}'

    missing_display_str = missing_display_str.lower()

    if missing_display_str == 'missing':
        missing_long_display_str = 'missing information'
    else:
        missing_long_display_str = missing_display_str


    below_text = f"""
    This table shows where the {count_with_missings_var} variable is {missing_long_display_str}. 
    For all panels, each item represents a subsample analysis where {row_byvar} is the value given by the row
    and where {col_byvar} is the value given by the column.
    Panel A describes the number of observations.
    Panel B describes the percentage of observations with {missing_long_display_str} for {count_with_missings_var}.
    Panel C describes the number of unique {id_var}s. 
    Panel D describes the percentage of unique {id_var}s which have more than {missing_tolerance} observations
    with {missing_long_display_str} for {count_with_missings_var}. 
    """ + extra_below_text

    if extra_caption:
        caption = f'{caption} - {extra_caption}'

    table = lt.Table.from_panel_name_df_dict(
        df_dict,
        include_index=True,
        top_left_corner_labels=col_byvar,
        below_text=below_text,
        caption=caption
    )

    if outfolder is not None:
        table.to_pdf_and_move(
            outname=caption,
            outfolder=outfolder
        )

    return table


from typing import Callable, List
import warnings

from dero.manager.selector.logic.get.main import get_dict_of_any_defined_pipeline_manager_names_and_instances
from dero.manager.logic.get import _get_from_nested_obj_by_section_path
from dero.manager.sectionpath.sectionpath import SectionPath
from dero.manager.pipelines.models.dictfile import PipelineDictFile
from dero.manager.data.models.dictfile import DataDictFile
from dero.manager.selector.logic.get.frommanager import get_pipeline_dict_path_and_data_dict_path_from_manager
from dero.manager.pipelines.models.collection import PipelineCollection
from dero.manager.basemodels.pipeline import Pipeline
from dero.manager.data.models.collection import DataCollection, DataSource
from dero.manager.views.object import ObjectView

class Selector:

    def __init__(self):
        self._attach_to_pipeline_manager()
        self._load_structure()
        self._is_selector = True

    def __contains__(self, item):
        if not isinstance(item, (str, SectionPath)):
            warnings.warn('check for if non-str object in Selector will always return False')
            return False

        collection_obj, relative_section_path = self._get_collection_obj_and_relative_section_path_from_structure(item)

        if relative_section_path is None:
            # got only the root data path, e.g. project.sources. Return the collection object itself
            return collection_obj

        # Only should return True if we find an ItemView
        # Three possible cases here.
        # Accessing a non-existent attr, should catch AttributeError then return False
        # Accessing an existing item, should be a collection, source, or function instance, return True
        # Accessing an existing attribute of an existing item, should not be an ItemView instance, return False
        try:
            result = _get_from_nested_obj_by_section_path(collection_obj, relative_section_path)
        except AttributeError:
            return False

        # TODO: converting all functions to pipelines would make this check safer
        item_types = (DataSource, DataCollection, Pipeline, PipelineCollection, Callable, ObjectView)
        if isinstance(result, item_types):
            return True

        return False


    def __getattr__(self, item):
        from dero.manager.selector.models.itemview import ItemView
        return ItemView(item, self)

    def __call__(self, item):
        from dero.manager.selector.models.itemview import ItemView
        return ItemView(item, self)

    def __dir__(self):
        exposed_methods = [
            'get_type'
        ]

        managers = list(self._managers.keys())

        return exposed_methods + managers

    def _get_dir_for_section_path(self, section_path_str: str) -> List[str]:
        collection_obj, relative_section_path = self._get_collection_obj_and_relative_section_path_from_structure(
            section_path_str
        )
        if relative_section_path is None:
            # got only the root data path, e.g. project.sources. Return the collection object itself
            return dir(collection_obj)

        result_obj = _get_from_nested_obj_by_section_path(collection_obj, relative_section_path)

        return dir(result_obj)

    def _get_collection_obj_and_relative_section_path_from_structure(self, section_path_str: str):
        # Handle accessing correct collection object.
        section_path = SectionPath(section_path_str)
        manager_name = section_path[0]

        if len(section_path) == 1:
            # Got only the manager, e.g. project
            # return only the manager itself
            return self._managers[manager_name], None


        if section_path[1] == 'sources':
            # got a data path
            if len(section_path) == 2:
                # got only the root data path, e.g. project.sources
                # return the collection object itself
                return self._structure[manager_name]['data'], None
            collection_name = 'data'
            section_path_begin_index = 2
        else:
            collection_name = 'funcs'
            section_path_begin_index = 1

        relative_section_path = SectionPath('.'.join(section_path[section_path_begin_index:]))
        collection_obj = self._structure[manager_name][collection_name]

        return collection_obj, relative_section_path

    def get_type(self, section_path_str: str):
        collection_obj, relative_section_path = self._get_collection_obj_and_relative_section_path_from_structure(
            section_path_str
        )

        if relative_section_path is None:
            # got only the root data path, e.g. project.sources. Return the collection object itself
            return collection_obj

        # TODO: nicer error than KeyError for typo
        result = _get_from_nested_obj_by_section_path(collection_obj, relative_section_path)

        return type(result)

    def _get_real_item(self, item):
        section_path = SectionPath(item)
        manager_name = section_path[0]
        manager = self._managers[manager_name]
        relative_section_path = SectionPath('.'.join(section_path[1:]))
        return _get_from_nested_obj_by_section_path(manager, relative_section_path)

    def _attach_to_pipeline_manager(self):
        self._managers = get_dict_of_any_defined_pipeline_manager_names_and_instances()

    def _load_structure(self):
        out_dict = {}
        for manager_name, manager in self._managers.items():
            pipeline_dict_path, data_dict_path = get_pipeline_dict_path_and_data_dict_path_from_manager(manager)
            pipeline_dict_file = PipelineDictFile(pipeline_dict_path, name='pipeline_dict')
            pipeline_dict = pipeline_dict_file.load()
            data_dict = DataDictFile(data_dict_path, name='data_dict').load()
            manager_dict = {
                'funcs': PipelineCollection.from_dict(
                    pipeline_dict,
                    manager.basepath,
                    imports=pipeline_dict_file.interface.imports
                ),
                'data': DataCollection.from_dict(data_dict, manager.basepath)
            }
            out_dict[manager_name] = manager_dict
        self._structure = out_dict




def _is_selector(obj) -> bool:
    is_selector = getattr(obj, '_is_selector', False)
    return is_selector




class EqOnAttrsMixin:
    equal_attrs = None

    def __eq__(self, other):
        if self.equal_attrs is None:
            return self == other

        for equal_attr in self.equal_attrs:
            if not hasattr(other, equal_attr): # other object doesn't have this property, must not be equal
                return False
            if getattr(self, equal_attr) != getattr(other, equal_attr):
                return False

        # passed all checks, must be equal
        return True


from typing import Union, List

from dero.latex.figure.models.subfigure import Subfigure, Graphic
from dero.latex.models.documentitem import DocumentItem
from dero.latex.models import Item
from dero.latex.models.caption import Caption
from dero.latex.models.label import Label
from dero.latex.models.landscape import Landscape
from dero.latex.logic.builder import build_figure_content
from dero.latex.texgen.replacements.filename import latex_filename_replacements

SubfigureOrGraphic = Union[Subfigure, Graphic]
SubfiguresOrGraphics = List[SubfigureOrGraphic]

class Figure(DocumentItem, Item):
    """
    used for creating latex figures from images. Currently the main usage is the Figure class created with the method
    Figure.from_dict_of_names_and_filepaths. Pass a dictionary where the keys are names for subfigures and the values
    are filepaths where the image for the subfigure is located.

    """
    name = 'figure'

    def __init__(self, subfigures: SubfiguresOrGraphics, caption=None, label=None, centering=True, position_str=None,
                 landscape: bool=False):
        self.subfigures = subfigures
        self.caption = Caption(caption) if caption else None
        self.label = Label(label) if label else None
        self.centering = centering
        self.landscape = landscape

        self._remove_subfigure_elevate_contents_to_figure_if_single_subfigure()

        content = build_figure_content(
            self.subfigures,
            caption=self.caption,
            label=self.label,
            centering=self.centering,
            position_str=position_str
        )

        super().__init__(self.name, content)

        # Landscape needs to go around figure tags, so wrap output rather than contents
        if landscape:
            self._output = Landscape().wrap(self._output)

    def __repr__(self):
        return f'<Figure(subfigures={self.subfigures}, caption={self.caption})>'

    def __iter__(self):
        for subfigure in self.subfigures:
            yield subfigure

    def __getitem__(self, item):
        return self.subfigures[item]

    def as_document(self, landscape=False):
        from dero.latex.models.document import Document
        from dero.latex.figure.packages import default_packages

        return Document(self, default_packages, landscape=landscape)

    def to_pdf_and_move(self, as_document=True, outfolder: str=None, outname: str=None,
                        landscape=False):
        from dero.latex.logic.pdf import _document_to_pdf_and_move
        from dero.latex.models.document import Document

        to_output: Figure = self
        if as_document:
            to_output: Document = self.as_document(
                landscape=landscape if self.landscape == False else False  # don't apply landscape twice
            )

        if outfolder is None:
            outfolder = '.'
        if outname is None:
            outname = 'figure'
        else:
            outname = latex_filename_replacements(outname)

        _document_to_pdf_and_move(
            to_output,
            outfolder,
            image_paths=self.filepaths,
            outname=outname,
            as_document=as_document
        )

    @property
    def filepaths(self):
        filepaths = []
        for subfigure in self:
            if isinstance(subfigure, Subfigure):
                filepaths.append(subfigure.graphic.filepath)
            elif isinstance(subfigure, Graphic):
                filepaths.append(subfigure.filepath)
            else:
                raise ValueError(f'must pass Subfigures or Graphics to Figure. got type {type(subfigure)}')
        return filepaths

    @property
    def source_paths(self):
        source_paths = []
        for subfigure in self:
            if isinstance(subfigure, Subfigure):
                source_paths.append(subfigure.graphic.source_path)
            elif isinstance(subfigure, Graphic):
                source_paths.append(subfigure.source_path)
            else:
                raise ValueError(f'must pass Subfigures or Graphics to Figure. got type {type(subfigure)}')
        return source_paths

    @classmethod
    def from_dict_of_names_and_filepaths(cls, filepath_name_dict: dict, figure_name: str=None,
                                         position_str_name_dict: dict=None, label=None, centering=True,
                                         landscape: bool = False):
        """

        Args:
            filepath_name_dict: dictionary where keys are names of subfigures and values
                                are the filepaths to the images for those subfigures.
            figure_name: name for overall figure
            position_str_name_dict: dictionary where keys are names of subfigures and values
                                are the position strs for those figures, e.g. r'[t]{0.45\linewidth}'

        Returns:

        """

        # TODO: add possibility of passing grid shape rather than actual position str

        if position_str_name_dict is None:
            position_str_name_dict = {}

        subfigures = []
        for name, filepath in filepath_name_dict.items():
            subfigures.append(
                Subfigure(
                    filepath,
                    caption=name,
                    position_str=position_str_name_dict[name] if name in position_str_name_dict else r'[t]{0.45\linewidth}'
                )
            )

        return cls(
            subfigures,
            caption=figure_name,
            label=label,
            centering=centering,
            landscape=landscape
        )

    def _remove_subfigure_elevate_contents_to_figure_if_single_subfigure(self):
        """
        If there is a single subfigure, leaving in subfigure format results in very small output. Must strip
        out subfigure layer, leaving only a figure with a graphic, then it can fill the page.
        :return:
        """
        if len(self.subfigures) != 1:
            return

        self.subfigures: List[Subfigure]
        # update width to whole page
        graphic = Graphic(self.subfigures[0].graphic.filepath, width=r'1.26\paperwidth')
        # need to turn off centering to cover whole page
        self.centering = False

        self.subfigures = [graphic]






from typing import Tuple

from dero.manager.main import PipelineManager


def get_pipeline_dict_path_and_data_dict_path_from_manager(manager: PipelineManager) -> Tuple[str, str]:
    return manager.pipeline_dict_path, manager.data_dict_path


from typing import Callable, Union, List, Dict, TYPE_CHECKING
if TYPE_CHECKING:
    from dero.manager.pipelines.models.collection import PipelineCollection

from dero.manager.basemodels.pipeline import Pipeline
from dero.manager.views.object import ObjectView

PipelineOrFunction = Union[Pipeline, Callable]
PipelineOrFunctionOrCollection = Union[PipelineOrFunction, 'PipelineCollection']
PipelinesOrFunctions = List[PipelineOrFunction]
PipelinesOrFunctionsOrCollections = List[PipelineOrFunctionOrCollection]
PipelineDict = Dict[str, Union[PipelinesOrFunctions, 'PipelineDict']]
PipelineDictOrPipelineOrFunction = Union[PipelineDict, PipelineOrFunction]
PipelineDictsOrPipelinesOrFunctions = List[PipelineDictOrPipelineOrFunction]
StrList = List[str]
ObjectViewOrCollection = Union[ObjectView, 'PipelineCollection']


import warnings

from dero.reg.differenced import _is_diff_reg_str
from dero.reg.quantile import _is_quantile_reg_str
from dero.reg.dataprep import _is_normal_reg_str
from dero.reg.linmodels.bindings.modelstr import _is_linearmodels_str
from dero.reg.linmodels.reg import linear_reg
from .differenced import diff_reg
from .quantile import quantile_reg
from .reg import reg
from dero.reg.lag.create import _set_lag_variables


def any_reg(reg_type, *reg_args, **reg_kwargs):

    reg_args, reg_kwargs = _validate_inputs(*reg_args, **reg_kwargs)

    reg_type = reg_type.lower()

    if _is_diff_reg_str(reg_type):
        temp_kwargs = reg_kwargs.copy()
        if 'diff_cols' in temp_kwargs:
            diff_cols = temp_kwargs.pop('diff_cols')
        else:
            diff_cols = None
        id_col = temp_kwargs.pop('id_col')
        date_col = temp_kwargs.pop('date_col')
        return diff_reg(*reg_args, id_col, date_col, diff_cols=diff_cols, **temp_kwargs)

    if _is_normal_reg_str(reg_type):
        return reg(*reg_args, **reg_kwargs)

    if _is_quantile_reg_str(reg_type):
        return quantile_reg(*reg_args, **reg_kwargs)

    if _is_linearmodels_str(reg_type):
        return linear_reg(*reg_args, **reg_kwargs)

    raise ValueError(f'Must pass valid reg type. Got {reg_type}')


def _validate_inputs(*args, **kwargs):
    yvar = args[1]
    xvars = args[2]

    # If an x variable is actually going to be lagged, don't need to exclude it if it matches the y variable
    if 'num_lags' in kwargs and kwargs['num_lags'] > 0:
        lagvars = _set_lag_variables(kwargs['lag_variables'] if 'lag_variables' in kwargs else [], yvar, xvars)
        if yvar in lagvars:
            # if y variable lagged, then remove any xvars that are also lagged
            # only consider those which are in lag variables
            examine_xvars = [var for var in lagvars if var != yvar]
        else:
            # y variable not lagged. shouldn't remove any xvars which are in lag variables
            # only consider those not in lag variables
            examine_xvars = [xvar for xvar in xvars if xvar not in lagvars]
    else:
        examine_xvars = xvars

    # Exclude x variable when matches y variable
    if yvar in examine_xvars:
        warnings.warn(f'{yvar} is both Y variable and passed in X variables. Removing from X for this model.', UserWarning)
        new_xvars = xvars.copy()
        new_xvars.remove(yvar)
        args = args[:2] + (new_xvars,) + args[3:]

    return args, kwargs



class StringAdditionMixin:

    def __add__(self, other):
        return str(self) + str(other)

    def __radd__(self, other):
        return str(other) + str(self)

    def join(self, iterable):
        return self.__str__().join(str(i) for i in iterable)





from typing import Union, Tuple, Dict

import pandas as pd

DfOrSeries = Union[pd.DataFrame, pd.Series]
PdDTypeQuadTuple = Tuple[DfOrSeries, DfOrSeries, DfOrSeries, DfOrSeries]
StrDict = Dict[str, str]


from typing import Callable

class SimplePropertyCacheMixin:

    def _try_getattr_else_call_func(self, attr: str, func: Callable, *func_args, **func_kwargs):
        try:
            return getattr(self, attr)
        except AttributeError:
            func(*func_args, **func_kwargs)

        return getattr(self, attr)


import pandas as pd
from typing import Tuple

from dero.reg.regtypes import InteractionTuples, StrOrListOfStrs
from .create import create_lagged_variables, _convert_interaction_tuples, _convert_variable_names, _set_lag_variables
from .remove import remove_lag_names_from_reg_results

DfYvarXvarInteractionTuplesLagvarTuple = Tuple[pd.DataFrame, str, StrOrListOfStrs, InteractionTuples, StrOrListOfStrs]

def create_lagged_variables_return_yvars_xvars_interaction_tuples(df: pd.DataFrame, yvar: str, xvars: StrOrListOfStrs,
                                                                  interaction_tuples: InteractionTuples=None,
                                                                  num_lags: int=0, lag_variables: StrOrListOfStrs='xvars',
                                                                  lag_period_var: str='Date', lag_id_var: str='TICKER',
                                                                  fill_method: str = 'ffill'
                                                                  ) -> DfYvarXvarInteractionTuplesLagvarTuple:
    if num_lags == 0:
        return df, yvar, xvars, interaction_tuples, lag_variables

    lag_variables = _set_lag_variables(lag_variables, yvar, xvars)
    df = create_lagged_variables(
        df,
        lag_variables,
        id_col=lag_id_var,
        date_col=lag_period_var,
        num_lags=num_lags,
        fill_method=fill_method
    )
    reg_yvar, reg_xvars = _convert_variable_names(yvar, xvars, lag_variables, num_lags=num_lags)
    if interaction_tuples is not None:
        interaction_tuples = _convert_interaction_tuples(interaction_tuples, lag_variables, num_lags=num_lags)

    return df, reg_yvar, reg_xvars, interaction_tuples, lag_variables


from typing import Callable, List, Union, Tuple
from copy import deepcopy
import pandas as pd
from functools import partial

from dero.manager.data.logic.merge.display import display_merge_summary
from dero.manager.data.models.source import DataSource
from dero.manager.data.logic.merge import left_merge_df
from dero.data.summarize import describe_df

StrList = List[str]
StrListOrNone = Union[None, StrList]
TwoDfTuple = Tuple[pd.DataFrame, pd.DataFrame]

class MergeOptions:

    def __init__(self, *merge_function_args, outpath=None, merge_function=left_merge_df,
                 left_df_keep_cols: StrListOrNone=None, right_df_keep_cols: StrListOrNone=None,
                 left_df_pre_process_func: Callable=None, right_df_pre_process_func: Callable=None,
                 left_df_pre_process_kwargs: dict=None, right_df_pre_process_kwargs: dict=None,
                 **merge_function_kwargs):
        """

        passed args to merge func will be (
            left_df_pre_process_func(left_df, **left_df_pre_process_kwargs)[left_df_keep_cols],
            right_df_pre_process_func(right_df, **right_df_pre_process_kwargs)[right_df_keep_cols],
            *merge_function_kwargs,
            **merge_function_kwargs
        )

        if left_df_keep_cols is None, will instead pass left_df_pre_process_func(left_df, **left_df_pre_process_kwargs).
        If right_df_keep_cols is None, will instead pass right_df_pre_process_func(right_df, **right_df_pre_process_kwargs).
        If left_df_pre_process_func is None, will instead pass left_df or left_df[left_df_keep_cols] depending
            on whether left_df_keep_cols was passed. Similar behavior for right.

        Args:
            *merge_function_args:
            outpath:
            merge_function:
            left_df_keep_cols:
            right_df_keep_cols:
            **merge_function_kwargs:
        """

        if left_df_pre_process_kwargs == None:
            left_df_pre_process_kwargs = {}

        if right_df_pre_process_kwargs == None:
            right_df_pre_process_kwargs = {}

        if left_df_pre_process_func == None:
            left_df_pre_process_func = lambda x: x

        if right_df_pre_process_func == None:
            right_df_pre_process_func = lambda x: x

        self.args = merge_function_args
        self.outpath = outpath
        self.merge_function = merge_function
        self.merge_function_kwargs = merge_function_kwargs
        self.left_df_keep_cols = left_df_keep_cols
        self.right_df_keep_cols = right_df_keep_cols
        self.left_df_pre_process_func = partial(left_df_pre_process_func, **left_df_pre_process_kwargs)
        self.right_df_pre_process_func = partial(right_df_pre_process_func, **right_df_pre_process_kwargs)


    def __repr__(self):
        return f'<DataMerge(args={self.args}, merge_function={self.merge_function.__name__}, ' \
               f'kwargs={self.merge_function_kwargs})>'

    def copy(self):
        return deepcopy(self)

    def update(self, **kwargs):
        self.merge_function_kwargs.update(**kwargs)


class DataMerge:

    def __init__(self, data_sources: [DataSource], merge_options: MergeOptions):
        self.data_sources = data_sources
        self.merge_options = merge_options
        self._merged_name = None
        self._merged_type = None
        self._merged_str = None
        self._set_result(merge_options.outpath)

    def merge(self):
        print(f'Running merge function {self.merge_str}')
        left_df, right_df = self._get_merge_dfs()
        self.result.df = self.merge_options.merge_function(
            left_df, right_df,
            *self.merge_options.args,
            **self.merge_options.merge_function_kwargs
        )
        print(f"""
        {self.data_sources[0].name_type} obs: {len(left_df)}
        {self.data_sources[1].name_type} obs: {len(right_df)}
        Merged obs: {len(self.result.df)}
        """)

    def summary(self, *summary_args, summary_method: str=None, summary_function: Callable=None,
                             summary_attr: str=None, **summary_method_kwargs):
        display_merge_summary(
            self,
            *summary_args,
            summary_method=summary_method,
            summary_function=summary_function,
            summary_attr=summary_attr,
            **summary_method_kwargs
        )

    def describe(self):
        display_merge_summary(
            self,
            summary_function=describe_df,
            disp=False # don't display from describe_df as will display from display_merge_summary
        )

    def _set_result(self, outpath=None):
        self.result = DataSource(outpath, name=self.merged_name, data_type=self.merged_type)

    def _get_merge_dfs(self) -> TwoDfTuple:
        left_df = self.data_sources[0].df
        right_df = self.data_sources[1].df

        # Handle pre process funcs
        if self.merge_options.left_df_pre_process_func is not None:
            left_df = self.merge_options.left_df_pre_process_func(left_df)
        if self.merge_options.right_df_pre_process_func is not None:
            right_df = self.merge_options.right_df_pre_process_func(right_df)

        # Handle selecting variables on processed df
        if self.merge_options.left_df_keep_cols is not None:
            left_df = left_df[self.merge_options.left_df_keep_cols]
        if self.merge_options.right_df_keep_cols is not None:
            right_df = right_df[self.merge_options.right_df_keep_cols]

        return left_df, right_df

    @property
    def merged_name(self):
        if self._merged_name is None:
            self._merged_name = self.data_sources[0].name + ' & ' + self.data_sources[1].name
        return self._merged_name

    @property
    def merged_type(self):
        if self._merged_type is None:
            self._merged_type = self.data_sources[0].data_type + ' & ' + self.data_sources[1].data_type
        return self._merged_type

    @property
    def merge_str(self):
        if self._merged_str is None:
            self._merged_str = f'''
            {self.merge_options.merge_function.__name__}(
                {self.data_sources[0].name_type},
                {self.data_sources[1].name_type},
                *{self.merge_options.args},
                **{self.merge_options.merge_function_kwargs}
            )
            '''
        return self._merged_str

    def __repr__(self):
        return f'<DataMerge(left={self.data_sources[0]}, right={self.data_sources[1]})>'


class LastMergeFinishedException(Exception):
    pass








class DataType:

    def __init__(self, name=None):
        self.name = name

    def __add__(self, other):
        return self.name + other

    def __radd__(self, other):
        return other + self.name

    def __repr__(self):
        return f'<DataType({self.name})>'

    def __eq__(self, other):
        # Handle when other is None
        if other is None:
            if self.name is None:
                return True
            else:
                return False

        # Handle when other is str
        if isinstance(other, str):
            return self.name == other

        # Handle when other is DataType
        return self.name == other.name



from dero.manager.selector.logic.get.fromglobs import get_dict_of_pipeline_manager_names_and_instances_from_globals
from dero.manager.selector.logic.get.globs import iter_globals

def get_dict_of_any_defined_pipeline_manager_names_and_instances():
    pm_dict = {}
    for globs in iter_globals():
        pm_dict.update(get_dict_of_pipeline_manager_names_and_instances_from_globals(globs=globs))

    return pm_dict


from typing import List

from dero.mixins.repr import ReprMixin
from dero.manager.basemodels.collection import Collection
from dero.manager.sectionpath.sectionpath import SectionPath
from dero.manager.logic.get import _get_from_nested_obj_by_section_path
from dero.manager.imports.models.statements.container import ImportStatementContainer

class Registrar(ReprMixin):
    repr_cols = ['name', 'basepath', 'collection']
    collection_class = Collection

    def __init__(self, collection: Collection, basepath: str, name=None):
        self.collection = collection
        self.basepath = basepath
        self.name = name

    def __getattr__(self, item):
        return getattr(self.collection, item)

    def __dir__(self):
        exposed_methods = [
            'scaffold_config',
            'get'
        ]
        exposed_attrs = [
            'basepath',
            'name'
        ]
        collection_attrs = [attr for attr in dir(self.collection) if attr not in exposed_methods + exposed_attrs]
        return exposed_methods + exposed_attrs + collection_attrs

    @classmethod
    def from_dict(cls, dict_: dict, basepath: str, name: str=None,
                  imports: ImportStatementContainer = None):
        collection = cls.collection_class.from_dict(
            dict_, basepath=basepath, name=name, imports=imports
        )

        return cls(collection, basepath=basepath, name=name)

    def scaffold_config(self):
        self.collection._output_config_files()

    def get(self, section_path_str: str):
        section_path = SectionPath(section_path_str)

        # Goes into nested sections, until it pulls the final section or pipeline
        return _get_from_nested_obj_by_section_path(self, section_path)


import pandas as pd

from dero.ext_pandas.optimize.typing import PdDTypeQuadTuple, StrDict


def optimized_df(df: pd.DataFrame) -> pd.DataFrame:
    columns = [col for col in df.columns]

    type_dfs: PdDTypeQuadTuple = _optimized_dtype_dfs(df)
    return pd.concat(type_dfs, axis=1)[columns]


def _optimized_dtype_dfs(df: pd.DataFrame) -> PdDTypeQuadTuple:
    obj_df = df.select_dtypes(include=['object'])
    if not obj_df.empty:
        obj_df = obj_df.astype('category')

    int_df = df.select_dtypes(include=['int'])
    if not int_df.empty:
        int_df = int_df.apply(pd.to_numeric, downcast='unsigned')

    float_df = df.select_dtypes(include=['float'])
    if not float_df.empty:
        float_df = float_df.apply(pd.to_numeric, downcast='float')

    type_dfs = (obj_df, int_df, float_df)

    optimized_columns = []
    for type_df in type_dfs:
        optimized_columns += [col for col in type_df.columns]

    # Excluded dtype should just be 'datetime', which does not need conversion
    excluded_columns = [col for col in df.columns if col not in optimized_columns]

    return type_dfs + (df[excluded_columns],)

def df_types_dict(df: pd.DataFrame, remove_dates=True) -> StrDict:
    df_types_dict = _df_types_dict(df)
    if remove_dates:
        return {col_name: dtype for col_name, dtype in df_types_dict.items() if 'date' not in dtype}
    else:
        return df_types_dict

def _df_types_dict(df: pd.DataFrame) -> StrDict:
    return df.dtypes.apply(lambda x: str(x)).to_dict()


import pandas as pd
from dero.latex import Document

from dero.data.typing import StrOrNone, DocumentOrLatexObjs
from dero.data.summarize.subset.missing.summary.graphs.pctbyid import missing_pct_by_id_figure
from dero.data.summarize.subset.missing.detail.main import obs_and_id_count_and_missing_pct_table

def missing_data_single_column_analysis(df: pd.DataFrame, col_with_missings: str, id_col: str,
                                        row_byvar: str, col_byvar: str,
                                        missing_tolerance: int=0,
                                        sort_cols_as_numeric: bool = True, sort_rows_as_numeric: bool = True,
                                        count_format_str: str = '.0f', pct_format_str: str = '.1f',
                                        missing_display_str: str = 'Missing',
                                        extra_caption: str='', table_extra_below_text: str='',
                                        outfolder: StrOrNone=None, as_document: bool=True,
                                        document_title: StrOrNone=None, author: StrOrNone=None,
                                        ) -> DocumentOrLatexObjs:

    missing_pct_fig = missing_pct_by_id_figure(
        df,
        id_col,
        col_with_missings,
        outfolder=outfolder
    )

    missing_pct_table = obs_and_id_count_and_missing_pct_table(
        df,
        col_with_missings,
        id_col,
        row_byvar,
        col_byvar,
        missing_tolerance=missing_tolerance,
        sort_cols_as_numeric=sort_cols_as_numeric,
        sort_rows_as_numeric=sort_rows_as_numeric,
        count_format_str=count_format_str,
        pct_format_str=pct_format_str,
        missing_display_str=missing_display_str,
        extra_caption=extra_caption,
        extra_below_text=table_extra_below_text,
        outfolder=outfolder
    )

    if not as_document:
        return [missing_pct_fig, missing_pct_table]

    if document_title is None:
        document_title = f'{missing_display_str} {col_with_missings} Analysis'

    if extra_caption:
        document_title = f'{document_title} - {extra_caption}'

    doc = Document(
        [missing_pct_fig, missing_pct_table],
        title=document_title,
        author=author
    )

    doc.to_pdf_and_move(
        outname=document_title,
        outfolder=outfolder
    )

    return doc


import pandas as pd
from dero.data.typing import StrList


def by_id_pct_long_df(df: pd.DataFrame, byvars: StrList, id_var: str,
                      count_with_missings_var: str, missing_tolerance: int = 0,
                      missing_display_str: str = 'Missing') -> pd.DataFrame:

    by_id_var = _by_id_count_long_df(
        df,
        byvars,
        id_var,
        count_with_missings_var,
        missing_tolerance=missing_tolerance,
        missing_display_str=missing_display_str
    )

    by_id_var[f'{id_var} Count'] = by_id_var[f'More than {missing_tolerance} {missing_display_str} {id_var}'] + \
                                   by_id_var[f'{missing_tolerance} or less {missing_display_str} {id_var} Count']

    by_id_var[f'More than {missing_tolerance} {missing_display_str} {id_var} Percentage'] = \
        (by_id_var[f'More than {missing_tolerance} {missing_display_str} {id_var}'] /
        by_id_var[f'{id_var} Count']) * 100

    by_id_var.drop([
        f'More than {missing_tolerance} {missing_display_str} {id_var}',
        f'{missing_tolerance} or less {missing_display_str} {id_var} Count'
    ], axis=1, inplace=True)

    return by_id_var


def _by_id_count_long_df(df: pd.DataFrame, byvars: StrList, id_var: str,
                         count_with_missings_var: str, missing_tolerance: int =0,
                         missing_display_str: str = 'Missing') -> pd.DataFrame:

    df['_one'] = 1  # temporary variable for counting without missing
    by_firm_counts = df.groupby([id_var] + byvars)[['_one', count_with_missings_var]].count().reset_index()
    df.drop('_one', axis=1, inplace=True)

    missing_df = by_firm_counts[by_firm_counts[count_with_missings_var] + missing_tolerance < by_firm_counts['_one']]
    full_df = by_firm_counts[by_firm_counts[count_with_missings_var] + missing_tolerance >= by_firm_counts['_one']]

    missing_counts = missing_df.groupby(byvars)[id_var].count()
    missing_counts.name = f'More than {missing_tolerance} {missing_display_str} {id_var}'

    full_counts = full_df.groupby(byvars)[id_var].count()
    full_counts.name = f'{missing_tolerance} or less {missing_display_str} {id_var} Count'

    by_id_var = pd.concat([missing_counts, full_counts], axis=1).fillna(0)
    return by_id_var.reset_index()


from dero.latex.models.package import Package

simple_package_strs = [
    'caption',
    'subcaption',
    'graphicx',
    'pdflscape'
]
simple_packages = [Package(str_) for str_ in simple_package_strs]

default_packages = simple_packages + []  # add any packages with options here with the Package class


from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from dero.manager.data.models.config import DataConfig

from dero.manager.basemodels.file import ConfigFileBase
from dero.manager.imports.models.statements.obj import ObjectImportStatement
from dero.manager.assignments.models.statement import AssignmentStatement
from dero.manager.data.models.source import DataSource
from dero.manager.io.file.interfaces.activeconfig import ActiveConfigFileInterface
from dero.manager.data.models.astitems import ast_dict_constructor_with_kwargs_from_dict

class DataConfigFile(ConfigFileBase):
    # lines to always import. pass import objects
    always_imports = [ObjectImportStatement.from_str('from dero.manager import Selector', preferred_position='begin')]

    # assignment lines to always include at beginning. pass assign objects
    always_assigns = [
        AssignmentStatement.from_str('s = Selector()', preferred_position='begin'),
        AssignmentStatement.from_str('loader_func_kwargs = dict(\n    \n)', preferred_position='end')
    ]

    def load(self, config_class: type = None) -> 'DataConfig':
        # Override base class method to pull a single dict, and not pass annotations
        from dero.manager.data.models.config import DataConfig

        # Data configs are a hybrid of the ast/static config and the active config
        self.active_interface = ActiveConfigFileInterface(self.interface.filepath)

        config_dict, annotation_dict = self.interface.load()
        user_defined_dict = self.active_interface.load()

        if config_class is None:
            config_class = DataConfig

        return config_class(
            d=config_dict,
            annotations=annotation_dict,
            active_config_dict=user_defined_dict,
            imports=self.interface.imports,
            _file=self,
            name=self.name
        )


from typing import List, Union, Tuple

from dero.latex.models.environment import Environment
from dero.latex.models import Item
from dero.latex.models.documentitem import DocumentItem
from dero.latex.texgen import _document_class_str
from dero.latex.models.package import Package
from dero.latex.texgen.packages import default_packages
from dero.latex.models.page.style import PageStyle
from dero.latex.models.landscape import Landscape
from dero.latex.logic.pdf import _document_to_pdf_and_move
from dero.latex.texgen.replacements.filename import latex_filename_replacements
from dero.latex.logic.extract.docitems import extract_document_items_from_ambiguous_collection
from dero.latex.models.page.number import right_aligned_page_numbers
from dero.latex.models.page.header import remove_header
from dero.latex.models.page.footer import CenterFooter

AnyItem = Union[Item, DocumentItem]
ListOfItems = List[AnyItem]
ItemOrListOfItems = Union[AnyItem, ListOfItems]
StrList = List[str]
StrListOrNone = Union[StrList, None]
ItemAndPreEnvContents = Tuple[AnyItem, StrListOrNone]

class DocumentEnvironment(Environment):
    name = 'document'

    def __init__(self):
        super().__init__(name=self.name)

class Document(DocumentItem, Item):
    name = 'document'

    def __init__(self, content: ItemOrListOfItems, packages: List[Package]=None, landscape=False,
                 title: str=None, author: str=None, date: str=None, skip_title_page: bool=False,
                 page_modifier_str: str='margin=0.8in, bottom=1.2in', page_header: bool=False,
                 page_numbers: bool=True):
        from dero.latex.logic.builder import _build
        from dero.latex.models.titlepage import TitlePage

        if packages is None:
            packages = default_packages.copy()

        # Set margins, body size, etc. with geometry package
        packages.append(Package('geometry', modifier_str=page_modifier_str))

        self.packages = packages

        possible_pre_env_contents = [
            _document_class_str(),
            *[str(package) for package in self.packages],
            PageStyle('fancy'),

            # header is there by default. add remove header lines if page_header=False
            remove_header if not page_header else None,

            # add right page numbers. if not, use blank center footer to clear default page numbers in center footer
            right_aligned_page_numbers if page_numbers else CenterFooter('')
        ]

        self.pre_env_contents = _build([item for item in possible_pre_env_contents if item is not None])

        if isinstance(content, Item):
            content = [content]

        if not skip_title_page and _should_create_title_page(title=title, author=author, date=date):
            title_page = TitlePage(title=title, author=author, date=date)
            content.insert(0, title_page)
            self.has_title_page = True
        else:
            self.has_title_page = False

        self.filepaths = self._get_filepaths_from_items(content)

        # combine content into a single str
        content = _build(content)

        if landscape:
            content = Landscape().wrap(str(content))

        super().__init__(self.name, content, pre_env_contents=self.pre_env_contents)

    def __repr__(self):
        return f'<Document>'

    def to_pdf_and_move(self, outfolder, outname='document',
                              move_folder_name='Tables', as_document=True):
        tex = str(self)

        outname = latex_filename_replacements(outname)

        _document_to_pdf_and_move(
            tex,
            outfolder=outfolder,
            outname=outname,
            image_paths=self.filepaths,
            move_folder_name=move_folder_name,
            as_document=as_document
        )

    def _get_filepaths_from_items(self, content: ListOfItems) -> StrListOrNone:
        from dero.latex.figure import Figure

        filepaths = []
        for item in content:
            # TODO: handling for other types which may have filepaths
            if isinstance(item, Figure):
                filepaths += item.filepaths

        if filepaths == []:
            return None

        return filepaths

    @classmethod
    def from_ambiguous_collection(cls, collection, packages: List[Package]=None, landscape=False,
                                  title=None, author=None, date=None, skip_title_page: bool=False,
                                  page_modifier_str: str = 'margin=0.8in, bottom=1.2in', page_header: bool = False,
                                  page_numbers: bool = True
                                  ):
        content = extract_document_items_from_ambiguous_collection(collection)

        return cls(content, packages=packages, landscape=landscape,
                   title=title, author=author, date=date, skip_title_page=skip_title_page,
                   page_modifier_str=page_modifier_str, page_header=page_header,
                   page_numbers=page_numbers)

def _should_create_title_page(title=None, author=None, date=None):
    return any([
        title is not None,
        author is not None,
        date is not None
    ])

def _content_items_and_collected_pre_env_contents():
    raise NotImplementedError('see TODO in _standardize_content_item_for_inclusion_in_document')

def _standardize_content_item_for_inclusion_in_document(item: AnyItem) -> ItemAndPreEnvContents:

    # No extra processing needed if not Document
    if not isinstance(item, Document):
        return item, None

    # TODO: restructure to remove title pages
    # TODO: restructure to extract content from Document

    return item, item.pre_env_contents




import pandas as pd

from dero.ext_pandas.optimize.typing import DfOrSeries


def read_file(filepath: str, **read_func_kwargs) -> DfOrSeries:
    extension = filepath.rpartition('.')[-1].lower()
    if extension == 'csv':
        df = pd.read_csv(filepath, **read_func_kwargs)
    # TODO: determine filetype and use proper loader
    else:
        raise NotImplementedError(f'could not load filetype {extension}')

    return df


from dero.latex.models.package import Package
from dero.latex.texgen.packages.columntypes import ColumnTypesPackage

_default_package_strs = [
    'amsmath',
    'pdflscape',
    'booktabs',
    'array',
    'threeparttable',
    'fancyhdr',
    'lastpage'
]

_direct_default_packages = [
    ColumnTypesPackage()
]

default_packages = [Package(package_str) for package_str in _default_package_strs] + _direct_default_packages





def _is_fama_macbeth_str(model_string: str) -> bool:
    simple_str = model_string.lower().strip().replace('-','').replace('_','').replace(' ','')
    return simple_str in ('famamacbeth', 'fmb', 'fb', 'fm', 'fmacbeth','famam','fmbreg','fmreg')

# Note: add model str functions here as they are created. This will feed into _is_linearmodels_str
model_str_funcs = [
    _is_fama_macbeth_str
]

def _is_linearmodels_str(model_string: str) -> bool:
    return any([is_model_str_func(model_string) for is_model_str_func in model_str_funcs])



from dero.mixins.repr import ReprMixin
from dero.mixins.attrequals import EqOnAttrsMixin
from dero.manager.selector.models.selector import Selector
from dero.manager.exceptions.pipelinemanager import PipelineManagerNotLoadedException
from copy import deepcopy


class ItemView(ReprMixin, EqOnAttrsMixin):
    """
    Class for representing a pipeline manager unit (function, pipeline, data source) without needing
    that unit to be loaded into pipeline manager. Allows using selector in config files and in app,
    by delaying looking up the item until an attribute/method is accessed or item is called.
    """
    repr_cols = ['section_path_str']
    equal_attrs = ['section_path_str']

    def __init__(self, section_path_str: str, selector: Selector):
        self.section_path_str = section_path_str
        self.selector = selector
        self._is_item_view = True

    def __getattr__(self, item):
        full_section_path_str = self.section_path_str + '.' + item
        if full_section_path_str in self.selector:
            # This is an item, return an item view for it
            return ItemView(full_section_path_str, self.selector)

        # Not an item.
        # Must be either an attribute of an item, or a typo. Actually look up the item now

        try:
            actual_item = self.item
        except PipelineManagerNotLoadedException:
            # Dealing with typos is difficult because if this is a typo and we are reaching here,
            # if PipelineManager.load() has not been run yet, we can't know the attributes of the items,
            # so there is no certain way to check whether this is an item attribute or a typo
            # TODO: check if manager is loaded, if so, then do a proper check. otherwise return a general error
            # TODO: saying that it could be either manager not being loaded or a typo and display part of the path not found
            raise PipelineManagerNotLoadedException


        return getattr(actual_item, item)

    def __dir__(self):
        exposed_properties = [
            'type',
            'item'
        ]
        collection_attrs = self.selector._get_dir_for_section_path(self.section_path_str)
        return exposed_properties + collection_attrs

    def __call__(self, *args, **kwargs):
        # When calling, assume user always wants the real item
        actual_item = self.selector._get_real_item(self.section_path_str)
        return actual_item(*args, **kwargs)

    def __deepcopy__(self, memodict={}):
        cls = self.__class__
        result = cls.__new__(cls)
        memodict[id(self)] = result
        deepcopy_skip_items = ['selector']
        deepcopy_dict = {key: value for key, value in self.__dict__.items() if key not in deepcopy_skip_items}
        shallow_copy_dict = {key: value for key, value in self.__dict__.items() if key in deepcopy_skip_items}
        for k, v in deepcopy_dict.items():
            setattr(result, k, deepcopy(v, memodict))
        result.__dict__.update(shallow_copy_dict)
        return result

    @property
    def type(self):
        return self.selector.get_type(self.section_path_str)

    @property
    def item(self):
        return self.selector._get_real_item(self.section_path_str)


def _is_item_view(obj) -> bool:
    is_item_view = getattr(obj, '_is_item_view', False)
    return is_item_view


from linearmodels.panel.results import PanelResults


class ModelData:

    def __init__(self, param_names):
        self.param_names = param_names


def _convert_linearmodels_result_to_statsmodels_result_format(result: PanelResults) -> None:
    """
    Note: inplace
    """

    # First at the result level
    move_dict = {
        'std_errors': 'bse',
        'tstats': 'tvalues'
    }
    for old_attr, new_attr in move_dict.items():
        _move_attr(result, old_attr, new_attr)

    # Now at the model level
    # Create result.model.data.param_names
    result.model.data = ModelData(
        param_names=[col for col in result.params.index],
    )
    result.model.endog_names = result.model.dependent.vars[0]


def _move_attr(obj: any, old_attr: str, new_attr: str) -> None:
    """
    Note: inplace
    """
    attr_value = getattr(obj, old_attr)
    setattr(obj, new_attr, attr_value)
    try:
        delattr(obj, old_attr)
    except AttributeError:
        pass


from dero.latex.texgen.replacements.file import general_latex_replacements

def _begin_str(str_):
    return rf'\begin{{{str_}}}'


def _end_str(str_):
    return fr'\end{{{str_}}}'


def _include_graphics_str(filepath, width=r'\linewidth'):
    return rf'\includegraphics[width={width}]{{{filepath}}}'

def _basic_item_str(item_name, contents):
    return rf'\{item_name}{{{general_latex_replacements(contents)}}}'

def _multi_option_item_str(item_name, *options):
    options_str = ''.join([f'{{{general_latex_replacements(str(option))}}}' for option in options])
    return rf'\{item_name}{options_str}'

def _cmidrule_str(align, col_str):
    return _multi_option_item_str(rf'cmidrule({align})', col_str)

def _centering_str():
    return r'\centering'

def _toprule_str():
    return r'\toprule'

def _midrule_str():
    return r'\midrule'

def _bottomrule_str():
    return r'\bottomrule'

def _document_class_str():
    return r'\documentclass{article}'

def _maketitle_str():
    return r'\maketitle'

def _todays_date_str():
    return r'\today'


def _usepackage_str(str_, modifier_str=None):
    if modifier_str:
        full_modifier_str = f'[{modifier_str}]'
    else:
        full_modifier_str = ''
    return rf'\usepackage{full_modifier_str}{{{str_}}}'





from dero.reg.models import get_model_class_by_string
from dero.reg.linmodels.bindings.input import _create_reg_df_y_x_and_lag_vars
from dero.reg.linmodels.bindings.fit import _estimate_handling_robust_and_cluster
from dero.reg.linmodels.bindings.result import _convert_linearmodels_result_to_statsmodels_result_format

def linear_reg(df, yvar, xvars, entity_var=None, time_var=None, robust=True, cluster=False, cons=True, fe=None, interaction_tuples=None,
        num_lags=0, lag_variables='xvars', lag_period_var='Date', lag_id_var='TICKER', lag_fill_method: str='ffill',
        model_type='fama macbeth', **fit_kwargs):
    """

    Args:
        df: pandas dataframe containing regression data
        yvar: str, column name of outcome y variable
        xvars: list of strs, column names of x variables for regression
        entity_var: str, name of variable identifying groups for panel
        time_var: str, name of variable identifying time for panel
        robust: bool, set to True to use heterskedasticity-robust standard errors
        cluster: False or str, set to a column name to calculate standard errors within clusters
                 given by unique values of given column name
        cons: bool, set to False to not include a constant in the regression
        fe: None or str or list of strs. If a str or list of strs is passed, uses these categorical
        variables to construct dummies for fixed effects.
        interaction_tuples: tuple or list of tuples of column names to interact and include as xvars
        num_lags: int, Number of periods to lag variables. Setting to other than 0 will activate lags
        lag_variables: 'all', 'xvars', or list of strs of names of columns to lag for regressions.
        lag_period_var: str, only used if lag_variables is not None. name of column which
                        contains period variable for lagging
        lag_id_var: str, only used if lag_variables is not None. name of column which
                        contains identifier variable for lagging
        lag_fill_method: str, 'ffill' or 'bfill' for which method to use to fill in missing rows when
                     creating lag variables. See pandas.DataFrame.fillna for more details
        model_type: str, 'fama macbeth' for type of model

    Returns:

    """

    if entity_var is None or time_var is None:
        raise ValueError('must pass both entity_var and time_var')


    regdf, y, X, lag_variables, dummy_cols_dict = _create_reg_df_y_x_and_lag_vars(
        df, yvar, xvars, entity_var, time_var,
        cluster=cluster,
        cons=cons, fe=fe,
        interaction_tuples=interaction_tuples,
        num_lags=num_lags,
        lag_variables=lag_variables,
        lag_period_var=lag_period_var,
        lag_id_var=lag_id_var,
        fill_method=lag_fill_method
    )

    ModelClass = get_model_class_by_string(model_type)
    mod = ModelClass(y, X)

    result = _estimate_handling_robust_and_cluster(regdf, mod, robust, cluster, **fit_kwargs)

    _convert_linearmodels_result_to_statsmodels_result_format(result)

    return result, dummy_cols_dict


